#
#   Sitestat 4.5/5.0 FTP library by GOD-EMPEROR
# 
#   core ftp support:
# 	FTP::Open <server> <user> <passwd> <?options?>
#	FTP::Close
#    	FTP::Cd <directory>
#	FTP::Pwd
#	FTP::Type <?ascii|binary?>	
#	FTP::List <?directory?>
#	FTP::NList <?directory?>
#	FTP::Size <file>
#	FTP::Delete <file>
#	FTP::Size <file>
#	FTP::Rename <from> <to>
#	FTP::Get <remote> <?local?>
#	FTP::Put <local> <?remote?>
#	FTP::Append <local> <?remote?>
#	FTP::MkDir <directory>
#	FTP::RmDir <directory>
#	FTP::Quote <arg1> <arg2> ...
#
######################################################################
# License, Original rights blurb
#
#   created:	12/96    changed:    04/99    version:    1.2
#
#   Copyright (C) 1996-1999 Steffen Traeger
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#   contact:
#	email:	Steffen.Traeger@t-online.de
#	url:	http://home.t-online.de/home/Steffen.Traeger
#
######################################################################

set FLVER "4.5.2"   ;## devel/upgrade started at 4.0 :)

package provide FTP $FLVER
namespace eval FTP {

namespace export DisplayMsg Open Close Cd Pwd Type List NList FileSize Delete Rename Quote MkDir RmDir 

set LOCALIP ""  ;##### set this if you have problems with it trying to open ports like 0,0,0,0,5,123 
set TIMEOUT 75  ;##### timeout in sec for ftp commands
set VERBOSE 1  ;##### show tons of debugging and protocol messages
set PASV 1  ;#####  set to 0 to open using active connection

#-----------------------------
# Fine tuning variable length timeouts

set fwait(login) 45
set fwait(user_sent) 45
set fwait(pass_sent) 20
set fwait(list2_open) 25
set fwait(list_open) 25
set fwait(list2_sent) 45
set fwait(list_sent) 45
set fwait(list2_close) 10
set fwait(list_close) 10
set fwait(delete_sent) 30
set fwait(rename_sent) 15
set fwait(rename_to) 15
set fwait(quote_sent) 30
set fwait(pwd_sent) 15
set fwait(type_sent) 15
set fwait(cd_sent) 15
set fwait(size_sent) 15
set fwait(mkdir_sent) 15
set fwait(rmdir_sent) 15
set fwait(chmod_sent) 20
set fwait(chown_sent) 20
set fwait(nuke_sent) 30
set fwait(noop_sent) 15
set fwait(quit_sent) 15
set fwait(delete) 10
set fwait(rename) 10
set fwait(rename) 10
set fwait(quote) 10
set fwait(pwd) 10
set fwait(type) 10
set fwait(cd) 10
set fwait(sizet) 10
set fwait(mkdir) 10
set fwait(rmdir) 10
set fwait(chmod) 10
set fwait(chown) 10
set fwait(nuke) 10
set fwait(noop) 10
set fwait(quit) 10


#############################################################################
#
# DisplayMsg --
#
# This is a simple procedure to display any messages on screen.
# It must be overwritten by users source code in the form:
# (exported)
#
#	namespace FTP {
#		proc DisplayMsg {msg} {
#			......
#		}
#	}
#
# Arguments:
# msg - 		message string
# state -		different states {normal, data, control, error}
#
proc DisplayMsg {msg {state ""}} {
 variable VERBOSE 
 variable FEB
 if {$state == "error"} {
  if ![info exists FEB] {set FEB 1} else {incr FEB}
  if {$FEB < 6} {
	putlog "ftp: ERROR : $msg"
  } elseif {$FEB == 6} {
	putlog "ftp: ERROR msg anti-flood - going silent"
  } elseif {$FEB > 25} {
	set FEB 1
	Close 1
  }
  return
 } elseif {$VERBOSE} {
  switch -- $state {
    data {putlog $msg}
    log {putlog "> $msg"}
    control {putlog $msg}
    default {putlog "ftp: $msg"}
  }
 }
set FEB 0
}

#############################################################################
#
# Timeout --
#
# Handle timeouts
# 
# Arguments:
#  -
#
proc Timeout {} {
 variable ftp
 variable fwait
 upvar #0 state state
 set state(control) 0
 DisplayMsg "control connection timeout after $fwait($ftp(State)) seconds" error
 Close 1
 unset ftp(State)
}

#############################################################################
#
# WaitOrTimeout --
#
# Blocks the running procedure and waits for a variable of the transaction 
# to complete. It continues processing procedure until a procedure or 
# StateHandler sets the value of variable "finished". 
# If a connection hangs the variable is setting instead of by this procedure after 
# specified seconds in $ftp(Timeout).
#  
# 
# Arguments:
#  -		
#

proc WaitOrTimeout {} {
 variable ftp
 variable fwait
 upvar #0 state state
 set retvar 1
 if {[info exists state(control)]} {
  catch {after cancel $ftp(Wait)}
  set ftp(Wait) [after [expr $fwait($ftp(State)) * 1000] [namespace current]::Timeout]
  vwait state(control)
 }
 catch {after cancel ftp(Wait)}
 return $state(control)
}

#############################################################################
#
# WaitComplete --
#
# Transaction completed.
# Cancel execution of the delayed command declared in procedure WaitOrTimeout.
# 
# Arguments:
# value -	result of the transaction
#			0 ... Error
#			1 ... OK
#

proc WaitComplete {value} {
 variable ftp
 variable fwait
 upvar #0 state state
 if {[info exists state(data)]} {
   catch {after cancel $ftp(Wait)}
   set ftp(Wait) [after [expr $fwait($ftp(State)) * 1000] Timeout]
   vwait state(data)
 }
 catch {after cancel $ftp(Wait)}
 set state(control) $value
}

#############################################################################
#
# PutsCtrlSocket --
#
# Puts then specified command to control socket,
# if DEBUG is set than it logs via DisplayMsg
# 
# Arguments:
# command - 		ftp command
#

proc PutsCtrlSock {{command ""}} {
 variable ftp 
 DisplayMsg "command -> $command" ; set reply ""
 catch {puts $ftp(CtrlSock) $command} reply
 if {[string match "*broken pipe*" $reply]} {
	DisplayMsg "broken pipe, socket fuckup" error
	Close 1
	return
 }
 flush $ftp(CtrlSock)
}

#############################################################################
#
# Close -- 
#
# Terminates a ftp session and if file transfer is not in progress, the server
# closes the control connection.  If file transfer is in progress, the 
# connection will remain open for result response and the server will then
# close it. 
# (exported)
# 
# Arguments:
# None.
#
# Returns:
# 0 -			ERROR
# 1 - 			OK
proc Close {{force 0}} {
  variable ftp

###  if {![info exists ftp(CtrlSock)]} {set ftp(CtrlSock) "xXxX"}
  if {$force == 1 || [lsearch [file channels] $ftp(CtrlSock)] == -1} {
	DisplayMsg "killing connection" control
	catch {close $ftp(CtrlSock)} 
	catch {unset ftp(CtrlSock)}
	catch {unset ftp(State)}
  } else {
	DisplayMsg "sending connection quit" control
	set ftp(State) quit
 	StateHandler
  } 
}

#############################################################################
#
# StateHandler --
#
# Implements a finite state handler and a fileevent handler
# for the control channel
# 
# Arguments:
# sock - 		socket name
#			If called from a procedure than this argument is empty.
# 			If called from a fileevent than this argument contains
#			the socket channel identifier.

proc StateHandler {{sock ""}} {
	upvar #0 state state
	variable ftp
	variable fwait
	variable VERBOSE
	variable TIMEOUT
	variable ato
	# there is no socket (and no channel to get) if called from a procedure
	set sc "000"
	if {$sock != ""} {
		set msgtext "" ; set bufline ""
		if {[lsearch [file channels] $ftp(CtrlSock)] == -1} {
			set sc 421
			DisplayMsg "service unavailable, remote host has closed the connection" error
			CloseDataConn
			WaitComplete 0
			return
		} else {
			if ![info exists ftp(multiline)] {set ftp(multiline) "-"}
			set number [gets $sock bufline]
			regexp "(^\[0-9\]+)( |-)?(.*)$" $bufline all sc ftp(multiline) msgtext
			if {$number > 1 && $sc != "000"} {
				DisplayMsg "$bufline" log
				if {$ftp(buffer) != ""} {append ftp(buffer) \n "$bufline"} else {set ftp(buffer) "$bufline"}
			}
			if {$ftp(multiline) == "-"} {return}
		}
	} else {
		DisplayMsg "StateHandler directly invoked" log
		set ato [clock clicks -milliseconds]
	}

	# use only the first digit 
	regexp "^\[0-9\]?" $sc rc
	DisplayMsg "code=$sc state=$ftp(State)" data
	if {$sc == "211" || $sc == "421"} {return 0}

 	switch -- $ftp(State) {
		login {
			switch $sc {
				220 {PutsCtrlSock "USER $ftp(User)" ; set ftp(State) user_sent}
			    default {set errmsg "unable to connect - $msgtext" ; set complete_with 0}
			}
		}
		user_sent {
			if {$rc == 3} {PutsCtrlSock "PASS $ftp(Passwd)" ; set ftp(State) pass_sent
			} else {set complete_with 0}
		}
		pass_sent {
			switch $rc {
			        2 {set complete_with 1}
			  default {set errmsg "unable to connect - $msgtext" ; set complete_with 0}
			}
		}

		list2_active	{
			if {[OpenActiveConn]} {
				PutsCtrlSock "PORT $ftp(LocalAddr),$ftp(DataPort)"
		  		set ftp(State) list2_open
			} else {
				set errmsg "unable to open port"
			}
		}
		list2_passive	{
			PutsCtrlSock "PASV"
			set ftp(State) list2_open
		}
		list2_open {
			switch $rc {
			    2  {
				if {$ftp(Mode) == "passive"} {
				     if {![OpenPassiveConn $ftp(buffer)]} {
				         set errmsg "unable to set PASV mode"
				       	 set complete_with 0
				     }
				}   
				PutsCtrlSock "LIST $ftp(Dir)"
				set ftp(State) list2_sent
			    }
			    default {
				if {$ftp(Mode) == "passive"} {
					set errmsg "unable to set PASV mode"
				} else {
					set errmsg "unable to open port"
				}  
				set complete_with 0
			    }
			}
		}
		list2_sent	{
			  switch $rc {
			    1       {
			               set ftp(State) list2_close
			            }
			    default {  
			               if { $ftp(Mode) == "passive" } {
			    	           unset state(data)
				       }    
				       set errmsg "unable to get directory listing"
				       set complete_with 0
			            }
			  }
		}
		list2_close {
			  switch $rc {
			    2     {
			                set complete_with 1
					set ftp(List) $ftp(buffer)
			            }
			    4     {
					set errmsg "bad list close"
			                set complete_with 0
					set ftp(List) $ftp(buffer)
			            }
			    default {
				       set errmsg "unable to receive list"
				       set complete_with 0
			            }
			  }
		}
		move	{
			PutsCtrlSock "SITE MOVE [lindex $ftp(Move) 0]"
			set ftp(State) move_sent
		}
		move_sent {
			if {$rc == 2} {	set complete_with 1; set ftp(Move) $ftp(buffer)} else {set complete_with 0}
		}
		nuke	{ 
			PutsCtrlSock "SITE NUKE [lindex $ftp(Nuke) 0]"
			set ftp(State) nuke_sent
		}
		nuke_sent {
			if {$rc == 2} {set complete_with 1; set ftp(Nuke) $ftp(buffer)} else {set complete_with 0}
		}
		noop	{ 
			PutsCtrlSock "NOOP"
			set ftp(State) noop_sent
		}
		noop_sent {
			if {$rc == 2} {set complete_with 1; set ftp(Noop) $ftp(buffer)} else {set complete_with 0}
		}
		chmod	{ 
			PutsCtrlSock "SITE CHMOD $ftp(Chmod1) [lindex $ftp(Chmod2) 0]"
			DisplayMsg ">--> CHMOD $ftp(Chmod1) [lindex $ftp(Chmod2) 0]"
			set ftp(State) chmod_sent
		}
		chmod_sent {
			set complete_with 1
			set ftp(Chmod1) "" ; set ftp(Chmod2) ""
		}
		chown	{ 
			PutsCtrlSock "SITE CHOWN $ftp(Chown1) [lindex $ftp(Chown2) 0]"
			DisplayMsg ">--> CHOWN $ftp(Chown1) [lindex $ftp(Chown2) 0]"
			set ftp(State) chown_sent
		}
		chown_sent {
			set complete_with 1
			set ftp(Chown1) "" ; set ftp(Chown2) ""
		}
		quit	{
		    	   PutsCtrlSock "QUIT"
			   set ftp(State) quit_sent
			}

		quit_sent {
			  switch $rc {
			    2       {
			               set complete_with 1
			            }
			    default {
				       set errmsg "unable to close connection - $msgtext"
				       set complete_with 0
			            }
			  }
			}
		
		quote	{
		    	   PutsCtrlSock [lindex $ftp(Cmd) 0]
			   set ftp(State) quote_sent
			}

		quote_sent {
	                   set complete_with 1
                           set ftp(Quote) $ftp(buffer)
			}
		
		type	{
		  	  if { $ftp(Type) == "ascii" } {
			  	PutsCtrlSock "TYPE A"
			  } elseif { $ftp(Type) == "binary" } {
			  	PutsCtrlSock "TYPE I"
			  }
  		  	  set ftp(State) type_sent
			}
			
		type_sent {
			  switch $rc {
			    2       {
				       set complete_with 1
			            }
			    default {
				       set errmsg "failed to set type - $ftp(Type)"
				       set complete_with 0
			            }
			  }
			}
	
		nlist_active {
			  if {[OpenActiveConn]} {
			    	PutsCtrlSock "PORT $ftp(LocalAddr),$ftp(DataPort)"
			  	set ftp(State) nlist_open
			  } else {
				set errmsg "unable to open port"
			  }
			  
		}
			
		nlist_passive {
		    PutsCtrlSock "PASV"
		    set ftp(State) nlist_open
		}
			
		nlist_open {
			  switch $rc {
			    2 {
			         if {$ftp(Mode) == "passive"} {
				     if ![OpenPassiveConn $ftp(buffer)] {
				          set errmsg "failed to set PASV mode"
				       	 set complete_with 0
				     }
				 }   
			         PutsCtrlSock "NLST $ftp(Dir)"
			  	 set ftp(State) list_sent
			      }
			    default {
			              if {$ftp(Mode) == "passive"} {
				          set errmsg "failed to set PASV mode"
				      } else {
				          set errmsg "unable to open port"
				      }  
			       	      set complete_with 0
			            }
			  }
		}
	
		list_active	{
			  if {[OpenActiveConn]} {
				PutsCtrlSock "PORT $ftp(LocalAddr),$ftp(DataPort)"
		  		set ftp(State) list_open
			  } else {
				set errmsg "unable to open port"
			  }
			  
		}
			
		list_passive	{
		    PutsCtrlSock "PASV"
		    set ftp(State) list_open
		}
			
		list_open {
			  switch $rc {
			    2  {
			         if {$ftp(Mode) == "passive"} {
				     if {![OpenPassiveConn $ftp(buffer)]} {
				          set errmsg "failed to set PASV mode"
				       	 set complete_with 0
				     }
				 }   
			         PutsCtrlSock "LIST $ftp(Dir)"
			  	 set ftp(State) list_sent
			       }
			    default {
			              if {$ftp(Mode) == "passive"} {
				          set errmsg "failed to set PASV mode"
				      } else {
				          set errmsg "unable to open port"
				      }  
				      set complete_with 0
			            }
			  }
		}
			
		list_sent	{
			  switch $rc {
			    1       {
			               set ftp(State) list_close
			            }
			    default {  
			               if { $ftp(Mode) == "passive" } {
			    	           unset state(data)
				       }    
				       set errmsg "unable to get directory listing"
				       set complete_with 0
			            }
			  }
		}

		list_close {
			  switch $rc {
			    2     {
			                set complete_with 1
			            }
			    4     {
					set errmsg "bad list close"
			                set complete_with 1
			            }
			    default {
				       set errmsg "unable to receive list"
				       set complete_with 0
			            }
			}
		}
									
		size {
			  PutsCtrlSock "SIZE $ftp(File)"
  		  	  set ftp(State) size_sent
			}

		size_sent {
			  switch $rc {
			    2       {
            			       regexp "^\[0-9\]+ (.*)$" $ftp(buffer) all ftp(FileSize)
			               set complete_with 1
			            }
			    default {
				       set errmsg "error getting file size"
				       set complete_with 0
			            }
			  }
			}


		pwd	{
			   PutsCtrlSock "PWD"
  		  	   set ftp(State) pwd_sent
			}
			
		pwd_sent {
			  switch $rc {
			    2       {
            			       regexp "^.*\"(.*)\"" $ftp(buffer) temp ftp(Dir)
			               set complete_with 1
			            }
			    default {
				       set errmsg "failed to get working dir"
				       set complete_with 0
			            }
			  }
			}

		cd	{
			   PutsCtrlSock "CWD [lindex $ftp(Dir) 0]"
  		  	   set ftp(State) cd_sent
			}
			
		cd_sent {
			  switch $rc {
			    2       {
			               set complete_with 1
			            }
			    default {
				       set errmsg "failed to change directory"
				       set complete_with 0
				     }
			  }
			}
			
		mkdir	{
			  PutsCtrlSock "MKD [lindex $ftp(Dir) 0]"
  		  	  set ftp(State) mkdir_sent
			}
			
		mkdir_sent {
			  switch $rc {
			    2       {
			               set complete_with 1
			            }
			    default {
				       set errmsg "failed to create dir - $ftp(Dir)"
				       set complete_with 0
				     }
			  }
			}
			
		rmdir	{
			  PutsCtrlSock "RMD [lindex $ftp(Dir) 0]"
  		  	  set ftp(State) rmdir_sent
			}
			
		rmdir_sent {
			  switch $rc {
			    2       {
			               set complete_with 1
			            }
			    default {
				       set errmsg "unable to remove directory"
				       set complete_with 0
				     }
			  }
			}
										
		delete	{
			  PutsCtrlSock "DELE [lindex $ftp(File) 0]"
  		  	  set ftp(State) delete_sent
			}
			
		delete_sent {
			  switch $rc {
			    2       {
			               set complete_with 1
			            }
			    default {
				       set errmsg "couldn't delete file - $ftp(File)"
				       set complete_with 0
				     }
			  }
			}
			
		rename	{
			  PutsCtrlSock "RNFR [lindex $ftp(RenameFrom) 0]"
  		  	  set ftp(State) rename_to
			}
			
		rename_to {
			  switch $rc {
			    3       {
			  	       PutsCtrlSock "RNTO [lindex $ftp(RenameTo) 0]"
  		  	  	       set ftp(State) rename_sent
			            }
			    default {
				       set errmsg "unable to rename file - [lindex $ftp(RenameFrom) 0]"
				       set complete_with 0
				     }
			  }
			}

		rename_sent {
			  switch $rc {
			    2     {
			               set complete_with 1
			            }
			    default {
				       set errmsg "failed to rename file - [lindex $ftp(RenameFrom) 0]"
				       set complete_with 0
				     }
			  }
			}
			
		put_active 	{
			  if {[OpenActiveConn]} {
			    	PutsCtrlSock "PORT $ftp(LocalAddr),$ftp(DataPort)"
			  	set ftp(State) put_open
			  } else {
				set errmsg "unable to open port"
			  }
			}
			
			
		put_passive	{
			               PutsCtrlSock "PASV"
			  	       set ftp(State) put_open
			}
			
		put_open {
			  switch $rc {
			    2  {
			         if {$ftp(Mode) == "passive"} {
				     if {![OpenPassiveConn $ftp(buffer)]} {
				         set errmsg "Error setting PASSIVE mode!"
				       	 set complete_with 0
				     }
				 }   
			         PutsCtrlSock "STOR $ftp(RemoteFilename)"
			         set ftp(State) put_sent
			       }
			    default {
			              if {$ftp(Mode) == "passive"} {
				          set errmsg "Error setting PASSIVE mode!"
				      } else {
				          set errmsg "Error setting port!"
				      }  
				      set complete_with 0
				    }
			  }
		}
			
		put_sent	{
			  switch $rc {
			    1       {
			               set ftp(State) put_close
			            }
			    default {
			              if {$ftp(Mode) == "passive"} {
			    	         # close already opened DataConnection
			    	         unset state(data)
				      }  
				       set errmsg "Error opening connection!"
				       set complete_with 0
				     }
			  }
		}
			
		put_close	{
			  switch $rc {
			    2       {
			  	       set complete_with 1
			            }
			    default {
				       set errmsg "Error storing file - $ftp(RemoteFilename)"
				       set complete_with 0
				     }
			  }
		}
			
		append_active 	{
			  if {[OpenActiveConn]} {
			    	PutsCtrlSock "PORT $ftp(LocalAddr),$ftp(DataPort)"
			  	set ftp(State) append_open
			  } else {
				set errmsg "Error setting port!"
			  }
			}
			
			
		append_passive	{
			               PutsCtrlSock "PASV"
			  	       set ftp(State) append_open
			}
			
		append_open {
			  switch $rc {
			    2  {
			         if {$ftp(Mode) == "passive"} {
				     if {![OpenPassiveConn $ftp(buffer)]} {
				         set errmsg "Error setting PASSIVE mode!"
				       	 set complete_with 0
				     }
				 }   
			         PutsCtrlSock "APPE $ftp(RemoteFilename)"
			         set ftp(State) append_sent
			       }
			    default {
			              if {$ftp(Mode) == "passive"} {
				          set errmsg "Error setting PASSIVE mode!"
				      } else {
				          set errmsg "Error setting port!"
				      }  
				      set complete_with 0
				    }
			  }
		}
			
		append_sent	{
			  switch $rc {
			    1       {
			               set ftp(State) append_close
			            }
			    default {
			              if {$ftp(Mode) == "passive"} {
			    	         # close already opened DataConnection
			    	         unset state(data)
				      }  
				       set errmsg "Error opening connection!"
				       set complete_with 0
				     }
			  }
		}
			
		append_close	{
			  switch $rc {
			    2       {
			  	       set complete_with 1
			            }
			    default {
				       set errmsg "Error storing file \"$ftp(RemoteFilename)\"!"
				       set complete_with 0
				     }
			  }
		}
			
		get_active 	{
			  if {[OpenActiveConn]} {
			    	PutsCtrlSock "PORT $ftp(LocalAddr),$ftp(DataPort)"
			  	set ftp(State) get_open
			  } else {
				set errmsg "Error setting port!"
			  }
			}
			
		get_passive {
			        PutsCtrlSock "PASV"
			  	set ftp(State) get_open
			    }
			
		get_open {
			  switch $rc {
			    2  -
			    3  {
			         if {$ftp(Mode) == "passive"} {
				     if {![OpenPassiveConn $ftp(buffer)]} {
				         set errmsg "Error setting PASSIVE mode!"
				       	 set complete_with 0
				     }
				 }   
			         PutsCtrlSock "RETR $ftp(RemoteFilename)"
			         set ftp(State) get_sent
			       }
			    default {
			              if {$ftp(Mode) == "passive"} {
				          set errmsg "Error setting PASSIVE mode!"
				      } else {
				          set errmsg "Error setting port!"
				      }  
				      set complete_with 0
				    }
			   }
			 }
			
		get_sent	{
			  switch $rc {
			    1 {
			         set ftp(State) get_close
			       }
			    default {
			              if {$ftp(Mode) == "passive"} {
			    	         # close already opened DataConnection
			    	         unset state(data)
				      }  
				      set errmsg "Error retrieving file \"$ftp(RemoteFilename)\"!"
				      set complete_with 0
				    }
			   }
		}
			
		get_close	{
			  switch $rc {
			    2       {
			  	       set complete_with 1
			            }
			    default {
				       set errmsg "Error retrieving file $ftp(RemoteFilename)"
				       set complete_with 0
				     }
			  }
		}
	}
	if {[array names fwait $ftp(State)] == ""} {set fwait($ftp(State)) $TIMEOUT}
	# finish waiting 
	if {[info exists complete_with]} {
		unset ftp(multiline); WaitComplete $complete_with
		if [info exists ato] {DisplayMsg "command duration -> [expr [clock clicks -milliseconds] - $ato]ms" log}
	}
	set ftp(buffer) ""
	if {[info exists errmsg]} {set ftp(Error) $errmsg ; DisplayMsg $errmsg error}
	DisplayMsg "StateHandler completed" log
}

proc Chmod {perms dir} {
	variable ftp 
	set ftp(State) chmod
	if ![connchk] {return 0}
	set ftp(Chmod1) $perms
	set ftp(Chmod2) [fixpath $dir]
	StateHandler
	set rc [WaitOrTimeout]
	return $rc
}
proc Chown {downer dir} {
	variable ftp 
	set ftp(State) chown
	if ![connchk] {return 0}
	set ftp(Chown1) $downer
	set ftp(Chown2) [fixpath $dir]
	StateHandler
	set rc [WaitOrTimeout]
	return $rc
}
proc Nuke {dir} {
	variable ftp 
	set ftp(State) nuke
	if ![connchk] {return 0}
	set ftp(Nuke) [fixpath $dir]
	StateHandler
	set rc [WaitOrTimeout]
	if {$rc} {return $ftp(Nuke)} else {return 0}
}
proc Noop {} {
	variable ftp 
	set ftp(State) noop
	if ![connchk] {return 0}
	StateHandler
	set rc [WaitOrTimeout]
	return $rc
}

proc Move {{dir ""}} {
	variable ftp 
	set ftp(State) move
	if ![connchk] {return 0}
	if { $dir == "" } {return}
	set ftp(Move) [fixpath $dir]
	StateHandler
	putlog "ftp: site move"
	set rc [WaitOrTimeout] 
	if {$rc} {return $ftp(Move)} else {return 0}
}

proc connchk {} {
  variable ftp
  variable CEB
  if {$CEB > 50} {set CEB 1} 
  if ![info exists ftp(CtrlSock)] {set ftp(CtrlSock) "xXxX"}
  if {[lsearch [file channels] $ftp(CtrlSock)] == -1} {
	set CEB 1
	if ![info exists ftp(State)] {set ftp(State) "no state"}
####	DisplayMsg "not connected ($ftp(State))" error
	return 0
  } else {
	incr CEB
	return 1
  }
}

proc fixpath {str} {
    return [list $str]
}


#############################################################################
#
# Type --
#
# REPRESENTATION TYPE - Sets the file transfer type to ascii or binary.
# (exported)
#
# Arguments:
# type - 		specifies the representation type (ascii|binary)
# 
# Returns:
# type	-		returns the current type or {} if an error occurs

proc Type {{type ""}} {
	variable ftp 
	# return current type
	set old_type $ftp(Type) 
	if { $type == "" || $type == $old_type} {return $ftp(Type)}
	
	set ftp(Type) $type
	set ftp(State) type
	if ![connchk] {return 0}

	StateHandler
	
	# wait for synchronization
	set rc [WaitOrTimeout]
	if {$rc} {
		return $ftp(Type)
	} else {
		# restore old type
		set ftp(Type) $old_type
		return {}
	}
 
}

#############################################################################
#
# NList --
#
# NAME LIST - This command causes a directory listing to be sent from
# server to user site.
# (exported)
# 
# Arguments:
# dir - 		The $dir should specify a directory or other system 
#			specific file group descriptor; a null argument 
#			implies the current directory. 
#
# Arguments:
# dir - 		directory to list 
# 
# Returns:
# sorted list of files or {} if listing fails

proc NList {{dir ""}} {
	variable ftp 

	set ftp(List) ""
	if { $dir == "" } {set ftp(Dir) "/"} else {set ftp(Dir) $dir}

	# save current type and force ascii mode
	set old_type $ftp(Type)
	if { $ftp(Type) != "ascii" } {Type ascii}

	set ftp(State) nlist_$ftp(Mode)
	if ![connchk] {return 0}

	StateHandler
	set rc [WaitOrTimeout]
	CloseDataConn

	# restore old type
	if { [Type] != $old_type } {Type $old_type}
	set retval $ftp(List)

	unset ftp(List)
	if {$rc} {return $retval} else {CloseDataConn ; return ""}
}

#############################################################################
#
# List --
#
# LIST - This command causes a list to be sent from the server
# to user site.
# (exported)
# 
# Arguments:
# dir - 		If the $dir specifies a directory or other group of 
#			files, the server should transfer a list of files in 
#			the specified directory. If the $dir specifies a file
#			then the server should send current information on the
# 			file.  A null argument implies the user's current 
#			working or default directory.  
# 
# Returns:
# list of files or {} if listing fails

proc List {mode {dir ""}} {
variable ftp 
variable PASV
	if {$dir == ""} {set ftp(Dir) "/"} else {set ftp(Dir) $dir}

	# save current type and force ascii mode
	set old_type $ftp(Type)
	if { $ftp(Type) != "ascii" } {Type ascii}

	set ftp(List) ""

	if {$mode == 0} {
		set ftp(State) list_$ftp(Mode)
	} else {
		set ftp(State) list2_$ftp(Mode)
	}
	if ![connchk] {return 0}

	StateHandler
	putlog "ftp: list ($mode)"
	set rc [WaitOrTimeout]

	if { [Type] != $old_type } {Type $old_type}
	CloseDataConn
	if {$rc} { 
		if {$PASV == 1} {
			set l [split $ftp(List) "\n"]
		} else {
			set l $ftp(List)
		}
		if {$mode == 0 && [lindex $l 0] == ""} {set l [lrange $l 1 end]} 
		return $l
	} else {
		return {}
	}
}


#############################################################################
#
# HandleList --
#
# Handles ascii data transfer for list commands
# 
# Arguments:
# sock - 		socket name (data channel)

proc HandleList0 {sock} {
	variable ftp 
	variable VERBOSE
	upvar #0 state state
	if {[lsearch [file channels] $sock] != -1} {
		set number [gets $sock bufline]
		if {$number > 1} {
			DisplayMsg "$bufline" log
			append ftp(List) "\{$bufline\} "
		}
		return
	} else {
		close $sock
   		unset state(data)
		if {$VERBOSE} {DisplayMsg "D: Port closed" data}
	}
}

proc HandleList1 {sock} {
	variable ftp 
	variable VERBOSE
	upvar #0 state state
	if ![eof $sock] {
		set buffer [read $sock]
		if { $buffer != "" } {
			append ftp(List) "$buffer"
		}	
	} else {
		close $sock
   		unset state(data)
		if {$VERBOSE} {DisplayMsg "D: Port closed" data}
	} 
}

#############################################################################
#
# File Size --
#
# REMOTE FILE SIZE - This command gets the file size of the
# file on the remote machine. 
# ATTANTION! Doesn't work properly in ascii mode!
# (exported)
# 
# Arguments:
# filename - 		specifies the remote file name
# 
# Returns:
# size -		files size in bytes or {} in error cases

proc Size {{filename ""}} {
variable ftp 

	if { $filename == "" } {return {}} 

	set ftp(File) $filename
	set ftp(FileSize) 0

	Type binary
	
	set ftp(State) size
	if ![connchk] {return 0}

	StateHandler
	set rc [WaitOrTimeout]

	set retval $ftp(FileSize)
	unset ftp(File)
	unset ftp(FileSize)

	if {$rc} {return $retval}
	return ""
}


#############################################################################
#
# Pwd --
#
# PRINT WORKING DIRECTORY - Causes the name of the current working directory.
# (exported)
# 
# Arguments:
# None.
# 
# Returns:
# current directory name

proc Pwd {} {
variable ftp 
	set ftp(Dir) ""
	set ftp(State) pwd
	if ![connchk] {return 0}

	StateHandler
	putlog "ftp: pwd"
	if {[WaitOrTimeout]} {return $ftp(Dir)}
	return ""
}

#############################################################################
#
# Cd --
#   
# CHANGE DIRECTORY - Sets the working directory on the server host.
# (exported)
# 
# Arguments:
# dir -			pathname specifying a directory
#
# Returns:
# 0 -			ERROR
# 1 - 			OK

proc Cd {{dir ""}} {
variable ftp 
	set ftp(State) cd
	if ![connchk] {return 0}
	if { $dir == "" } {
		set ftp(Dir) ""
	} else {
		set ftp(Dir) [fixpath $dir]
	}
	StateHandler
	set rc [WaitOrTimeout] 
	unset ftp(Dir)
	if {$rc} {return 1}
	return 0
}

#############################################################################
#
# MkDir --
#
# MAKE DIRECTORY - This command causes the directory specified in the $dir
# to be created as a directory (if the $dir is absolute) or as a subdirectory
# of the current working directory (if the $dir is relative).
# (exported)
# 
# Arguments:
# dir -			new directory name
#
# Returns:
# 0 -			ERROR
# 1 - 			OK

proc MkDir {dir} {
variable ftp 
	set ftp(Dir) [fixpath $dir]
	set ftp(State) mkdir
	if ![connchk] {return 0}
	StateHandler
	# wait for synchronization
	set rc [WaitOrTimeout] 
	unset ftp(Dir)
	if {$rc} {return 1}
	return 0
}

#############################################################################
#
# RmDir --
#
# REMOVE DIRECTORY - This command causes the directory specified in $dir to 
# be removed as a directory (if the $dir is absolute) or as a 
# subdirectory of the current working directory (if the $dir is relative).
# (exported)
#
# Arguments:
# dir -			directory name
#
# Returns:
# 0 -			ERROR
# 1 - 			OK

proc RmDir {dir} {
variable ftp 

	set ftp(Dir) [fixpath $dir]
	set ftp(State) rmdir

	if ![connchk] {return 0}
	StateHandler
	# wait for synchronization
	set rc [WaitOrTimeout] 
	unset ftp(Dir)
	if {$rc} {return 1}
	return 0
}

#############################################################################
#
# Delete --
#
# DELETE - This command causes the file specified in $file to be deleted at 
# the server site.
# (exported)
# 
# Arguments:
# file -			file name
#
# Returns:
# 0 -			ERROR
# 1 - 			OK

proc Delete {file} {
variable ftp 
	set ftp(File) [fixpath $file]
	set ftp(State) delete

	if ![connchk] {return 0}
	StateHandler
	# wait for synchronization
	set rc [WaitOrTimeout] 
	unset ftp(File)
	if {$rc} {return 1}
	return 0
}

#############################################################################
#
# Rename --
#
# RENAME FROM TO - This command causes the file specified in $from to be 
# renamed at the server site.
# (exported)
# 
# Arguments:
# from -			specifies the old file name of the file which 
#				is to be renamed
# to -				specifies the new file name of the file 
#				specified in the $from agument
# Returns:
# 0 -			ERROR
# 1 - 			OK

proc Rename {from to} {
	variable ftp 
	set ftp(RenameFrom) [fixpath $from]
	set ftp(RenameTo) [fixpath $to]
	set ftp(State) rename

	if ![connchk] {return 0}

	StateHandler
	# wait for synchronization
	set rc [WaitOrTimeout] 

	unset ftp(RenameFrom)
	unset ftp(RenameTo)
	if {$rc} {return 1}
	return 0
}

#############################################################################
#
# Quote -- 
#
# The arguments specified are sent, verbatim, to the remote FTP server.     
#
# Arguments:
# 	arg1 arg2 ...
#
# Returns:
#  string sent back by the remote FTP server or null string if any error
#

proc Quote {args} {
	variable ftp
	set ftp(State) quote
	if ![connchk] {return 0}
	set ftp(Quote) ""
	set ftp(Cmd) [list $args]

	StateHandler
	if {[string match -nocase [lindex $args 0] "site"]} {putlog "ftp: site [tlow [lindex $args 1]]"}

	set rc [WaitOrTimeout] 
	unset ftp(Cmd)
	if {$rc} {
		return $ftp(Quote)
	} else {
		return {}
	}
}


#############################################################################
#
# Open --
#
# Starts the ftp session and sets up a ftp control connection.
# (exported)
# 
# Arguments:
# server - 		ip number or hostname of server
# port -		port number of server
# user -		username
# passwd -		user's password
# options -		-Blocksize size	writes "size" bytes at once
#						(default 4096)
#			-Timeout seconds	if non-zero, sets up timeout to
#						occur after specified number of
#						seconds (default 120)
#			-Progress proc	procedure name that handles callbacks
#						(no default)  
#			-Mode mode		switch active or passive file transfer
#						(default active)
# 
# Returns:
# 0 -			Not logged in
# 1 - 		User logged in

proc Open {server port user passwd {args ""}} {
variable ftp
variable PASV
variable VERBOSE
variable TIMEOUT
upvar #0 state state
variable CEB
	if ![info exists CEB] {set CEB 1} 
	if ![info exists VERBOSE] {set VERBOSE 0}
##putlog "oooo [Noop]"
	if {$CEB > 20} {DisplayMsg "remaining disconnected" ; return 0}
	set ftp(State) 		login
	if [connchk] {DisplayMsg "already connected" ; return 1}

	set ftp(User) 		$user
	set ftp(Passwd) 	$passwd
	set ftp(RemoteHost) 	$server
	set ftp(Port) 		$port
	set ftp(LocalHost) 	[info hostname]
	set ftp(DataPort) 	0
	set ftp(Type) 		{}
	set ftp(Error) 		{}
	set ftp(Progress) 	{}
	set ftp(Blocksize) 	1024
	set ftp(buffer) ""

	if {$PASV} {
		set ftp(Mode) passive
	} else {
		set ftp(Mode) active
	}

	# set state var
	set state(control) 0 
	set opts "Starting new connection with: "

	# Get and set possible options
	set options {-Blocksize -Timeout -Mode -Progress}
	if {$args == ""} {
		append opts "no options"
	} else {
		foreach {option value} $args {
			if { [lsearch -exact $options $option] != "-1" } {
				append opts " $option=$value"
				set ftp($option) $value
			} 
		}
	}
	DisplayMsg $opts log

	# No call of StateHandler is required at this time.
	# StateHandler at first time is called automatically
	# by a fileevent for the control channel.

	# Try to open a control connection
	if ![OpenControlConn] {return 0}

	# waits for synchronization
	#   0 ... Not logged in
	#   1 ... User logged in
	if {[WaitOrTimeout]} {
		DisplayMsg "ftp connection opened" log
		return 1
	} else {
		DisplayMsg "failed to open ftp connection" log
		Close 1
		return 0
	}
}


############################################################################
#
# CloseDataConn -- 
#
# Closes all sockets and files used by the data conection
#
# Arguments:
# None.
#
# Returns:
# None.
#
proc CloseDataConn {} {
variable ftp
	DisplayMsg "closing all data connections"
	catch {after cancel $ftp(Wait)}
	catch {close $ftp(DataSock); unset ftp(DataSock)}
	catch {close $ftp(DestCI); unset ftp(DestCI)} 
	catch {close $ftp(SourceCI); unset ftp(SourceCI)}
####	catch {close $ftp(DummySock); unset ftp(DummySock)}
}

#############################################################################
#
# InitDataConn --
#
# Configures new data channel for connection to ftp server 
# ATTENTION! The new data channel "sock" is not the same as the 
# server channel, it's a dummy.
# 
# Arguments:
# sock -		the name of the new channel
# addr -		the address, in network address notation, 
#			of the client's host,
# port -		the client's port number

proc InitDataConn {sock addr port} {
variable ftp
upvar #0 state state
variable PASV

	# If the new channel is accepted, the dummy channel will be closed
	catch {close $ftp(DummySock); unset ftp(DummySock)}
	set state(data) ""

	# Configure translation mode
	if { $ftp(Type) == "ascii" } {
		fconfigure $sock -buffering line -blocking 0
	} else {
		fconfigure $sock -buffering line -translation binary -blocking 1
	}

	# assign fileevent handlers, source and destination CI (Channel Identifier)
	switch -regexp $ftp(State) {

		list {
			  fileevent $sock readable [list [namespace current]::HandleList$PASV $sock]
			  set ftp(SourceCI) $sock		  
			}

		get	{
			  fileevent $sock readable [list [namespace current]::HandleData $sock]
			  set ftp(SourceCI) $sock			  
			}

		append  -
		
		put {
			  fileevent $sock writable [list [namespace current]::HandleData $sock]
			  set ftp(DestCI) $sock			  
			}
	}
	DisplayMsg "D: Connection from $addr:$port" data
}

#############################################################################
#
# OpenActiveConn --
#
# Opens a ftp data connection
# 
# Arguments:
# None.
# 
# Returns:
# 0 -			no connection
# 1 - 			connection established

proc OpenActiveConn {} {
variable ftp
variable VERBOSE
variable LOCALIP

	# Port address 0 is a dummy used to give the server the responsibility 
	# of getting free new port addresses for every data transfer.
	set rc [catch {set ftp(DummySock) [socket -server [namespace current]::InitDataConn 0]} msg]
	if { $rc != 0 } {
       		DisplayMsg "ACTIVE LIST - $msg" error
       		return 0
		catch {close $ftp(DummySock)}
	}

	# get a new local port address for data transfer and convert it to a format
	# which is useable by the PORT command
	if {$LOCALIP != ""} {
		regsub -all "\[.\]" $LOCALIP "," ftp(LocalAddr) 
	}
	set p [lindex [fconfigure $ftp(DummySock) -sockname] 2]
	if {$VERBOSE} {
		DisplayMsg "D: Port is $p" data
	}
	set ftp(DataPort) [expr "$p / 256"],[expr "$p % 256"]

	return 1
}

#############################################################################
# OpenPassiveConn --
# Arguments: buffer - returned line from server control connection 
# 
proc OpenPassiveConn {buffer} {
variable ftp
	if {[regexp {([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+)} $buffer all a1 a2 a3 a4 p1 p2]} {

		set ftp(LocalAddr) "$a1.$a2.$a3.$a4"
		set ftp(DataPort) "[expr $p1 * 256 + $p2]"

		# establish data connection for passive mode
		catch {set ftp(DataSock) [socket -async $ftp(LocalAddr) $ftp(DataPort)]} msg

#		after 1000 [set rc [fconfigure $ftp(DataSock) -error]]
#		if { $rc != 0 } {
#			DisplayMsg "PASV LIST - $msg" error
#			return 0
#			catch {close $ftp(DataSock)}
#		}
		InitDataConn $ftp(DataSock) $ftp(LocalAddr) $ftp(DataPort)
		return 1
	} else {
		return 0
	}
}
#############################################################################
#
# OpenControlConn --
#
# Opens a ftp control connection
# 
# Arguments:
# None.
# 
# Returns:
# 0 -			no connection
# 1 - 			connection established

proc OpenControlConn {} {
variable ftp
variable VERBOSE
variable LOCALIP
	# open a control channel
	set sc ""
	DisplayMsg "opening connection to server -> $ftp(RemoteHost)" log
        catch {socket -async $ftp(RemoteHost) $ftp(Port)} sc
	set rc ""
	DisplayMsg ":: [list $sc] ::" control
	if [string match "sock*" $sc] {
		after 1800 {set rcdelay 1} ; vwait rcdelay
		set rc [fconfigure $sc -error]
	}
	if {$rc != ""} {
		DisplayMsg "not connected to server -> $rc" error
		catch {close $sc}
		unset ftp(State)
       		return 0
	}
	set ftp(CtrlSock) $sc

	DisplayMsg "socket connected -> $sc" log

	# configure control channel
	fconfigure $ftp(CtrlSock) -buffering line -blocking 0 -translation {auto crlf}
        fileevent $ftp(CtrlSock) readable [list [namespace current]::StateHandler $ftp(CtrlSock)]
	
	# prepare local ip address for PORT command (convert pointed format to comma format)
	if {$LOCALIP != ""} {
		set ftp(LocalAddr) $LOCALIP
	} else {
		set ftp(LocalAddr) [lindex [fconfigure $ftp(CtrlSock) -sockname] 0]
	}
	regsub -all "\[.\]" $ftp(LocalAddr) "," ftp(LocalAddr) 
	return 1
}
####if {![info exists ftp(CtrlSock)]} {set ftp(CtrlSock) xXxX}


# not forgotten close-brace (end of namespace)
}
